from .benchmark_math.ArchimedesOptimizationAlgorithm import AOA
from .benchmark_math.EelGrouperOptimizer import EGO
from .benchmark_math.ExplorationEnchancedGreyWolfOptimizer import EEGWO
from .benchmark_math.GeneticAlgorithm import GA
from .benchmark_math.GreaterCaneRatAlgorithmOctave import GCRA as GCRAOctave
from .benchmark_math.GreaterCaneRatAlgorithm import GCRA
from .benchmark_math.GreyWolfOptimizer import GWO
from .benchmark_math.InspiredGreyWolfOptimizer import IGWO
from .benchmark_math.KomodoMlipirAlgorithm import KMA
from .benchmark_math.MemorybasedGreyWolfOptimizer import mGWO
from .benchmark_math.ParticleSwarmOptimization import PSOClassic
from .benchmark_math.SlimeMouldAlgorithm import SMA
from .benchmark_math.WalrusOptimizationAlgorithm import WaOA
from .benchmark_math.WhaleOptimizationAlgorithm import WOA
