from .ArchimedesOptimizationAlgorithm import AOA
from .EelGrouperOptimizer import EGO
from .ExplorationEnchancedGreyWolfOptimizer import EEGWO
from .GeneticAlgorithm import GA
from .GreaterCaneRatAlgorithm import GCRA
from .GreyWolfOptimizer import GWO
from .InspiredGreyWolfOptimizer import IGWO
from .KomodoMlipirAlgorithm import KMA
from .MemorybasedGreyWolfOptimizer import mGWO
from .ParticleSwarmOptimization import PSOClassic
from .SlimeMouldAlgorithm import SMA
from .WalrusOptimizationAlgorithm import WaOA
from .WhaleOptimizationAlgorithm import WOA
from .MuskOxOptimizer import MO
